@@ -1,9 +1,9 @@
 #!/bin/bash
-if [ "$(uname)" == "Darwin" ]
+if [ "$(uname)" = "Darwin" ]
 then
     # for Mac OSX
     export KERAS_BACKEND=tensorflow
-elif [ "$(uname)" == "Linux" ]
+elif [ "$(uname)" = "Linux" ]
 then
     # for Linux
     export KERAS_BACKEND=theano
