@@ -13,6 +13,7 @@ from . import backend as K
 from . import optimizers
 from . import layers as layer_module
 from .utils.io_utils import ask_to_proceed_with_overwrite
+from .utils.generic_utils import has_arg
 from .engine.training import Model
 from .engine import topology
 from .engine.topology import Layer
@@ -98,77 +99,82 @@ def save_model(model, filepath, overwrite=True, include_optimizer=True):
         if not proceed:
             return
 
-    f = h5py.File(filepath, 'w')
-    f.attrs['keras_version'] = str(keras_version).encode('utf8')
-    f.attrs['backend'] = K.backend().encode('utf8')
-    f.attrs['model_config'] = json.dumps({
-        'class_name': model.__class__.__name__,
-        'config': model.get_config()
-    }, default=get_json_type).encode('utf8')
+    with h5py.File(filepath, mode='w') as f:
+        f.attrs['keras_version'] = str(keras_version).encode('utf8')
+        f.attrs['backend'] = K.backend().encode('utf8')
+        f.attrs['model_config'] = json.dumps({
+            'class_name': model.__class__.__name__,
+            'config': model.get_config()
+        }, default=get_json_type).encode('utf8')
 
-    model_weights_group = f.create_group('model_weights')
-    if legacy_models.needs_legacy_support(model):
-        model_layers = legacy_models.legacy_sequential_layers(model)
-    else:
-        model_layers = model.layers
-    topology.save_weights_to_hdf5_group(model_weights_group, model_layers)
-
-    if include_optimizer and hasattr(model, 'optimizer'):
-        if isinstance(model.optimizer, optimizers.TFOptimizer):
-            warnings.warn(
-                'TensorFlow optimizers do not '
-                'make it possible to access '
-                'optimizer attributes or optimizer state '
-                'after instantiation. '
-                'As a result, we cannot save the optimizer '
-                'as part of the model save file.'
-                'You will have to compile your model again after loading it. '
-                'Prefer using a Keras optimizer instead '
-                '(see keras.io/optimizers).')
+        model_weights_group = f.create_group('model_weights')
+        if legacy_models.needs_legacy_support(model):
+            model_layers = legacy_models.legacy_sequential_layers(model)
         else:
-            f.attrs['training_config'] = json.dumps({
-                'optimizer_config': {
-                    'class_name': model.optimizer.__class__.__name__,
-                    'config': model.optimizer.get_config()
-                },
-                'loss': model.loss,
-                'metrics': model.metrics,
-                'sample_weight_mode': model.sample_weight_mode,
-                'loss_weights': model.loss_weights,
-            }, default=get_json_type).encode('utf8')
+            model_layers = model.layers
+        topology.save_weights_to_hdf5_group(model_weights_group, model_layers)
 
-            # Save optimizer weights.
-            symbolic_weights = getattr(model.optimizer, 'weights')
-            if symbolic_weights:
-                optimizer_weights_group = f.create_group('optimizer_weights')
-                weight_values = K.batch_get_value(symbolic_weights)
-                weight_names = []
-                for i, (w, val) in enumerate(zip(symbolic_weights, weight_values)):
-                    # Default values of symbolic_weights is /variable for theano and cntk
-                    if K.backend() == 'theano' or K.backend() == 'cntk':
-                        if hasattr(w, 'name') and w.name != "/variable":
-                            name = str(w.name)
+        if include_optimizer and hasattr(model, 'optimizer'):
+            if isinstance(model.optimizer, optimizers.TFOptimizer):
+                warnings.warn(
+                    'TensorFlow optimizers do not '
+                    'make it possible to access '
+                    'optimizer attributes or optimizer state '
+                    'after instantiation. '
+                    'As a result, we cannot save the optimizer '
+                    'as part of the model save file.'
+                    'You will have to compile your model again '
+                    'after loading it. '
+                    'Prefer using a Keras optimizer instead '
+                    '(see keras.io/optimizers).')
+            else:
+                f.attrs['training_config'] = json.dumps({
+                    'optimizer_config': {
+                        'class_name': model.optimizer.__class__.__name__,
+                        'config': model.optimizer.get_config()
+                    },
+                    'loss': model.loss,
+                    'metrics': model.metrics,
+                    'sample_weight_mode': model.sample_weight_mode,
+                    'loss_weights': model.loss_weights,
+                }, default=get_json_type).encode('utf8')
+
+                # Save optimizer weights.
+                symbolic_weights = getattr(model.optimizer, 'weights')
+                if symbolic_weights:
+                    optimizer_weights_group = f.create_group('optimizer_weights')
+                    weight_values = K.batch_get_value(symbolic_weights)
+                    weight_names = []
+                    for i, (w, val) in enumerate(zip(symbolic_weights,
+                                                     weight_values)):
+                        # Default values of symbolic_weights is /variable
+                        # for theano and cntk
+                        if K.backend() == 'theano' or K.backend() == 'cntk':
+                            if hasattr(w, 'name'):
+                                if w.name.split('/')[-1] == 'variable':
+                                    name = str(w.name) + '_' + str(i)
+                                else:
+                                    name = str(w.name)
+                            else:
+                                name = 'param_' + str(i)
                         else:
-                            name = 'param_' + str(i)
-                    else:
-                        if hasattr(w, 'name') and w.name:
-                            name = str(w.name)
+                            if hasattr(w, 'name') and w.name:
+                                name = str(w.name)
+                            else:
+                                name = 'param_' + str(i)
+                        weight_names.append(name.encode('utf8'))
+                    optimizer_weights_group.attrs['weight_names'] = weight_names
+                    for name, val in zip(weight_names, weight_values):
+                        param_dset = optimizer_weights_group.create_dataset(
+                            name,
+                            val.shape,
+                            dtype=val.dtype)
+                        if not val.shape:
+                            # scalar
+                            param_dset[()] = val
                         else:
-                            name = 'param_' + str(i)
-                    weight_names.append(name.encode('utf8'))
-                optimizer_weights_group.attrs['weight_names'] = weight_names
-                for name, val in zip(weight_names, weight_values):
-                    param_dset = optimizer_weights_group.create_dataset(
-                        name,
-                        val.shape,
-                        dtype=val.dtype)
-                    if not val.shape:
-                        # scalar
-                        param_dset[()] = val
-                    else:
-                        param_dset[:] = val
-    f.flush()
-    f.close()
+                            param_dset[:] = val
+        f.flush()
 
 
 def load_model(filepath, custom_objects=None, compile=True):
@@ -516,12 +522,12 @@ class Sequential(Model):
         # Returns
             A layer instance.
         """
-        if self.model is None:
+        if not self.built:
             self.build()
         return self.model.get_layer(name, index)
 
     def call(self, inputs, mask=None):
-        if self.model is None:
+        if not self.built:
             self.build()
         return self.model.call(inputs, mask)
 
@@ -555,12 +561,11 @@ class Sequential(Model):
         # Make sure child model callbacks
         # will call the parent Sequential model.
         self.model.callback_model = self
-
         self.built = True
 
     @property
     def uses_learning_phase(self):
-        if self.model is None:
+        if not self.built:
             self.build()
         return self.model.uses_learning_phase
 
@@ -625,44 +630,38 @@ class Sequential(Model):
 
     @property
     def updates(self):
-        if self.model is None:
+        if not self.built:
             self.build()
         return self.model.updates
 
     @property
     def state_updates(self):
-        if self.model is None:
+        if not self.built:
             self.build()
         return self.model.state_updates
 
     def get_updates_for(self, inputs):
-        if self.model is None:
+        if not self.built:
             self.build()
         return self.model.get_updates_for(inputs)
 
     @property
     def losses(self):
-        if self.model is None:
+        if not self.built:
             self.build()
         return self.model.losses
 
     def get_losses_for(self, inputs):
-        if self.model is None:
+        if not self.built:
             self.build()
         return self.model.get_losses_for(inputs)
 
     @property
     def regularizers(self):
-        if self.model is None:
+        if not self.built:
             self.build()
         return self.model.regularizers
 
-    @property
-    def constraints(self):
-        if self.model is None:
-            self.build()
-        return self.model.constraints
-
     def get_weights(self):
         """Retrieves the weights of the model.
 
@@ -678,7 +677,7 @@ class Sequential(Model):
                 weights.append(layer.get_weights())
             return weights
 
-        if self.model is None:
+        if not self.built:
             self.build()
         return self.model.get_weights()
 
@@ -698,7 +697,7 @@ class Sequential(Model):
                 layer.set_weights(weights[:nb_param])
                 weights = weights[nb_param:]
 
-        if self.model is None:
+        if not self.built:
             self.build()
         self.model.set_weights(weights)
 
@@ -743,24 +742,48 @@ class Sequential(Model):
     def compile(self, optimizer, loss,
                 metrics=None,
                 sample_weight_mode=None,
+                weighted_metrics=None,
+                target_tensors=None,
                 **kwargs):
-        """Configures the learning process.
+        """Configures the model for training.
 
         # Arguments
-            optimizer: str (name of optimizer) or optimizer object.
+            optimizer: String (name of optimizer) or optimizer object.
                 See [optimizers](/optimizers).
-            loss: str (name of objective function) or objective function.
+            loss: String (name of objective function) or objective function.
                 See [losses](/losses).
-            metrics: list of metrics to be evaluated by the model
+                If the model has multiple outputs, you can use a different loss
+                on each output by passing a dictionary or a list of losses.
+                The loss value that will be minimized by the model
+                will then be the sum of all individual losses.
+            metrics: List of metrics to be evaluated by the model
                 during training and testing.
                 Typically you will use `metrics=['accuracy']`.
-                See [metrics](/metrics).
-            sample_weight_mode: if you need to do timestep-wise
-                sample weighting (2D weights), set this to "temporal".
-                "None" defaults to sample-wise weights (1D).
-            **kwargs: for Theano/CNTK backends, these are passed into
-                K.function. When using the TensorFlow backend, these are
-                passed into `tf.Session.run`.
+                To specify different metrics for different outputs of a
+                multi-output model, you could also pass a dictionary,
+                such as `metrics={'output_a': 'accuracy'}`.
+            sample_weight_mode: If you need to do timestep-wise
+                sample weighting (2D weights), set this to `"temporal"`.
+                `None` defaults to sample-wise weights (1D).
+                If the model has multiple outputs, you can use a different
+                `sample_weight_mode` on each output by passing a
+                dictionary or a list of modes.
+            weighted_metrics: List of metrics to be evaluated and weighted
+                by sample_weight or class_weight during training and testing.
+            target_tensors: By default, Keras will create placeholders for the
+                model's target, which will be fed with the target data during
+                training. If instead you would like to use your own
+                target tensors (in turn, Keras will not expect external
+                Numpy data for these targets at training time), you
+                can specify them via the `target_tensors` argument. It can be
+                a single tensor (for a single-output model), a list of tensors,
+                or a dict mapping output names to target tensors.
+            **kwargs: When using the Theano/CNTK backends, these arguments
+                are passed into K.function. When using the TensorFlow backend,
+                these arguments are passed into `tf.Session.run`.
+
+        # Raises
+            ValueError: In case of invalid arguments for
 
         # Example
             ```python
@@ -778,17 +801,20 @@ class Sequential(Model):
         self.model.compile(optimizer, loss,
                            metrics=metrics,
                            sample_weight_mode=sample_weight_mode,
+                           weighted_metrics=weighted_metrics,
+                           target_tensors=target_tensors,
                            **kwargs)
         self.optimizer = self.model.optimizer
         self.loss = self.model.loss
-        self.total_loss = self.model.total_loss
-        self.loss_weights = self.model.loss_weights
         self.metrics = self.model.metrics
+        self.loss_weights = self.model.loss_weights
+        self.sample_weight_mode = self.model.sample_weight_mode
+        self.weighted_metrics = self.model.weighted_metrics
+        self.targets = self.model.targets
         self.metrics_tensors = self.model.metrics_tensors
         self.metrics_names = self.model.metrics_names
-        self.sample_weight_mode = self.model.sample_weight_mode
         self.sample_weights = self.model.sample_weights
-        self.targets = self.model.targets
+        self.total_loss = self.model.total_loss
 
     def fit(self, x, y, batch_size=32, epochs=10, verbose=1, callbacks=None,
             validation_split=0., validation_data=None, shuffle=True,
@@ -800,7 +826,11 @@ class Sequential(Model):
                 (if the model has multiple inputs).
             y: labels, as a Numpy array.
             batch_size: integer. Number of samples per gradient update.
-            epochs: integer, the number of epochs to train the model.
+            epochs: integer. Number of epochs to train the model.
+                Note that in conjunction with initial_epoch, the parameter
+                epochs is to be understood as "final epoch". The model is
+                not trained for a number of steps given by epochs, but
+                until the epoch epochs is reached.
             verbose: 0 for no logging to stdout,
                 1 for progress bar logging, 2 for one log line per epoch.
             callbacks: list of `keras.callbacks.Callback` instances.
@@ -827,8 +857,8 @@ class Sequential(Model):
                 to apply a different weight to every timestep of every sample.
                 In this case you should make sure to specify
                 sample_weight_mode="temporal" in compile().
-            initial_epoch: epoch at which to start training
-                (useful for resuming a previous training run)
+            initial_epoch: Epoch at which to start training
+                (useful for resuming a previous training run).
 
         # Returns
             A `History` object. Its `History.history` attribute is
@@ -847,7 +877,7 @@ class Sequential(Model):
         if kwargs:
             raise TypeError('Unrecognized keyword arguments: ' + str(kwargs))
 
-        if self.model is None:
+        if not self.built:
             raise RuntimeError('The model needs to be compiled '
                                'before being used.')
         return self.model.fit(x, y,
@@ -883,7 +913,7 @@ class Sequential(Model):
         # Raises
             RuntimeError: if the model was never compiled.
         """
-        if self.model is None:
+        if not self.built:
             raise RuntimeError('The model needs to be compiled '
                                'before being used.')
         return self.model.evaluate(x, y,
@@ -904,7 +934,7 @@ class Sequential(Model):
         # Returns
             A Numpy array of predictions.
         """
-        if self.model is None:
+        if not self.built:
             self.build()
         return self.model.predict(x, batch_size=batch_size, verbose=verbose)
 
@@ -918,7 +948,7 @@ class Sequential(Model):
         # Returns
             A Numpy array of predictions.
         """
-        if self.model is None:
+        if not self.built:
             self.build()
         return self.model.predict_on_batch(x)
 
@@ -943,7 +973,7 @@ class Sequential(Model):
         # Raises
             RuntimeError: if the model was never compiled.
         """
-        if self.model is None:
+        if not self.built:
             raise RuntimeError('The model needs to be compiled '
                                'before being used.')
         return self.model.train_on_batch(x, y,
@@ -969,7 +999,7 @@ class Sequential(Model):
         # Raises
             RuntimeError: if the model was never compiled.
         """
-        if self.model is None:
+        if not self.built:
             raise RuntimeError('The model needs to be compiled '
                                'before being used.')
         return self.model.test_on_batch(x, y,
@@ -1029,6 +1059,7 @@ class Sequential(Model):
                       max_queue_size=10,
                       workers=1,
                       use_multiprocessing=False,
+                      shuffle=True,
                       initial_epoch=0):
         """Fits the model on data generated batch-by-batch by a Python generator.
 
@@ -1051,6 +1082,10 @@ class Sequential(Model):
                 be equal to the number of unique samples of your dataset
                 divided by the batch size.
             epochs: Integer, total number of iterations on the data.
+                Note that in conjunction with initial_epoch, the parameter
+                epochs is to be understood as "final epoch". The model is
+                not trained for n steps given by epochs, but until the
+                epoch epochs is reached.
             verbose: Verbosity mode, 0, 1, or 2.
             callbacks: List of callbacks to be called during training.
             validation_data: This can be either
@@ -1074,8 +1109,11 @@ class Sequential(Model):
                 non picklable arguments to the generator
                 as they can't be passed
                 easily to children processes.
+            shuffle: Whether to shuffle the order of the batches at
+                the beginning of each epoch. Only used with instances
+                of `Sequence` (keras.utils.Sequence).
             initial_epoch: Epoch at which to start training
-                (useful for resuming a previous training run)
+                (useful for resuming a previous training run).
 
         # Returns
             A `History` object.
@@ -1094,13 +1132,13 @@ class Sequential(Model):
                         # and labels, from each line in the file
                         x, y = process_line(line)
                         yield (x, y)
-                        f.close()
+                    f.close()
 
             model.fit_generator(generate_arrays_from_file('/my_file.txt'),
                                 steps_per_epoch=1000, epochs=10)
         ```
         """
-        if self.model is None:
+        if not self.built:
             raise RuntimeError('The model needs to be compiled '
                                'before being used.')
         return self.model.fit_generator(generator,
@@ -1114,6 +1152,7 @@ class Sequential(Model):
                                         max_queue_size=max_queue_size,
                                         workers=workers,
                                         use_multiprocessing=use_multiprocessing,
+                                        shuffle=shuffle,
                                         initial_epoch=initial_epoch)
 
     @interfaces.legacy_generator_methods_support
@@ -1147,7 +1186,7 @@ class Sequential(Model):
         # Raises
             RuntimeError: if the model was never compiled.
         """
-        if self.model is None:
+        if not self.built:
             raise RuntimeError('The model needs to be compiled '
                                'before being used.')
         return self.model.evaluate_generator(generator,
@@ -1181,7 +1220,7 @@ class Sequential(Model):
         # Returns
             A Numpy array of predictions.
         """
-        if self.model is None:
+        if not self.built:
             self.build()
         return self.model.predict_generator(generator, steps,
                                             max_queue_size=max_queue_size,
@@ -1289,3 +1328,231 @@ class Sequential(Model):
             layer = get_or_create_layer(conf)
             model.add(layer)
         return model
+
+
+def _clone_functional_model(model, input_tensors=None):
+    """Clone a functional `Model` instance.
+
+    Model cloning is similar to calling a model on new inputs,
+    except that it creates new layers (and thus new weights) instead
+    of sharing the weights of the existing layers.
+
+    # Arguments
+        model: Instance of `Model`.
+        input_tensors: optional list of input tensors
+            to build the model upon. If not provided,
+            placeholders will be created.
+
+    # Returns
+        An instance of `Model` reproducing the behavior
+        of the original model, on top of new inputs tensors,
+        using newly instantiated weights.
+
+    # Raises
+        ValueError: in case of invalid `model` argument value.
+    """
+    if not isinstance(model, Model):
+        raise ValueError('Expected `model` argument '
+                         'to be a `Model` instance, got ', model)
+    if isinstance(model, Sequential):
+        raise ValueError('Expected `model` argument '
+                         'to be a functional `Model` instance, '
+                         'got a `Sequential` instance instead:', model)
+
+    layer_map = {}  # Cache for created layers.
+    tensor_map = {}  # Map {reference_tensor: (corresponding_tensor, mask)}
+    if input_tensors is None:
+        # Create placeholders to build the model on top of.
+        input_layers = []
+        input_tensors = []
+        for layer in model.input_layers:
+            input_tensor = Input(batch_shape=layer.batch_input_shape,
+                                 dtype=layer.dtype,
+                                 sparse=layer.sparse,
+                                 name=layer.name)
+            input_tensors.append(input_tensor)
+            # Cache newly created input layer.
+            newly_created_input_layer = input_tensor._keras_history[0]
+            layer_map[layer] = newly_created_input_layer
+        for original_input_layer, cloned_input_layer in zip(model.input_layers, input_layers):
+            layer_map[original_input_layer] = cloned_input_layer
+    else:
+        # Make sure that all input tensors come from a Keras layer.
+        # If tensor comes from an input layer: cache the input layer.
+        input_tensors = topology._to_list(input_tensors)
+        _input_tensors = []
+        for i, x in enumerate(input_tensors):
+            if not K.is_keras_tensor(x):
+                name = model.input_layers[i].name
+                input_tensor = Input(tensor=x,
+                                     name='input_wrapper_for_' + name)
+                _input_tensors.append(input_tensor)
+                # Cache newly created input layer.
+                original_input_layer = x._keras_history[0]
+                newly_created_input_layer = input_tensor._keras_history[0]
+                layer_map[original_input_layer] = newly_created_input_layer
+            else:
+                _input_tensors.append(x)
+        input_tensors = _input_tensors
+
+    for x, y in zip(model.inputs, input_tensors):
+        tensor_map[x] = (y, None)  # tensor, mask
+
+    # Iterated over every node in the reference model, in depth order.
+    depth_keys = list(model.nodes_by_depth.keys())
+    depth_keys.sort(reverse=True)
+    for depth in depth_keys:
+        nodes = model.nodes_by_depth[depth]
+        for node in nodes:
+            # Recover the corresponding layer.
+            layer = node.outbound_layer
+
+            # Get or create layer.
+            if layer not in layer_map:
+                # Clone layer.
+                new_layer = layer.__class__.from_config(layer.get_config())
+                layer_map[layer] = new_layer
+                layer = new_layer
+            else:
+                # Reuse previously cloned layer.
+                layer = layer_map[layer]
+                # Don't call InputLayer multiple times.
+                if isinstance(layer, topology.InputLayer):
+                    continue
+
+            # Gather inputs to call the new layer.
+            reference_input_tensors = node.input_tensors
+            reference_output_tensors = node.output_tensors
+
+            # If all previous input tensors are available in tensor_map,
+            # then call node.inbound_layer on them.
+            computed_data = []  # List of tuples (input, mask).
+            for x in reference_input_tensors:
+                if x in tensor_map:
+                    computed_data.append(tensor_map[x])
+
+            if len(computed_data) == len(reference_input_tensors):
+                # Call layer.
+                if node.arguments:
+                    kwargs = node.arguments
+                else:
+                    kwargs = {}
+                if len(computed_data) == 1:
+                    computed_tensor, computed_mask = computed_data[0]
+                    if has_arg(layer.call, 'mask'):
+                        if 'mask' not in kwargs:
+                            kwargs['mask'] = computed_mask
+                    output_tensors = topology._to_list(
+                        layer(computed_tensor, **kwargs))
+                    output_masks = topology._to_list(
+                        layer.compute_mask(computed_tensor,
+                                           computed_mask))
+                    computed_tensors = [computed_tensor]
+                    computed_masks = [computed_mask]
+                else:
+                    computed_tensors = [x[0] for x in computed_data]
+                    computed_masks = [x[1] for x in computed_data]
+                    if has_arg(layer.call, 'mask'):
+                        if 'mask' not in kwargs:
+                            kwargs['mask'] = computed_masks
+                    output_tensors = topology._to_list(
+                        layer(computed_tensors, **kwargs))
+                    output_masks = topology._to_list(
+                        layer.compute_mask(computed_tensors,
+                                           computed_masks))
+                # Update tensor_map.
+                for x, y, mask in zip(reference_output_tensors,
+                                      output_tensors,
+                                      output_masks):
+                    tensor_map[x] = (y, mask)
+
+    # Check that we did compute the model outputs,
+    # then instantiate a new model from inputs and outputs.
+    output_tensors = []
+    for x in model.outputs:
+        assert x in tensor_map, 'Could not compute output ' + str(x)
+        tensor, _ = tensor_map[x]
+        output_tensors.append(tensor)
+    return Model(input_tensors, output_tensors, name=model.name)
+
+
+def _clone_sequential_model(model, input_tensors=None):
+    """Clone a `Sequential` model instance.
+
+    Model cloning is similar to calling a model on new inputs,
+    except that it creates new layers (and thus new weights) instead
+    of sharing the weights of the existing layers.
+
+    # Arguments
+        model: Instance of `Sequential`.
+        input_tensors: optional list of input tensors
+            to build the model upon. If not provided,
+            placeholders will be created.
+
+    # Returns
+        An instance of `Sequential` reproducing the behavior
+        of the original model, on top of new inputs tensors,
+        using newly instantiated weights.
+
+    # Raises
+        ValueError: in case of invalid `model` argument value.
+    """
+    if not isinstance(model, Sequential):
+        raise ValueError('Expected `model` argument '
+                         'to be a `Sequential` model instance, '
+                         'but got:', model)
+
+    def clone(layer):
+        return layer.__class__.from_config(layer.get_config())
+
+    layers = [clone(layer) for layer in model.layers]
+    if input_tensors is None:
+        return Sequential(layers=layers, name=model.name)
+    else:
+        if len(topology._to_list(input_tensors)) != 1:
+            raise ValueError('To clone a `Sequential` model, we expect '
+                             ' at most one tensor '
+                             'as part of `input_tensors`.')
+        x = topology._to_list(input_tensors)[0]
+        if K.is_keras_tensor(x):
+            origin_layer = x._keras_history[0]
+            if isinstance(origin_layer, topology.InputLayer):
+                return Sequential(layers=[origin_layer] + layers,
+                                  name=model.name)
+            else:
+                raise ValueError('Cannot clone a `Sequential` model on top '
+                                 'of a tensor that comes from a Keras layer '
+                                 'other than an `InputLayer`. '
+                                 'Use the functional API instead.')
+        input_tensor = Input(tensor=x,
+                             name='input_wrapper_for_' + str(x.name))
+        input_layer = input_tensor._keras_history[0]
+        return Sequential(layers=[input_layer] + layers, name=model.name)
+
+
+def clone_model(model, input_tensors=None):
+    """Clone any `Model` instance.
+
+    Model cloning is similar to calling a model on new inputs,
+    except that it creates new layers (and thus new weights) instead
+    of sharing the weights of the existing layers.
+
+    # Arguments
+        model: Instance of `Model`
+            (could be a functional model or a Sequential model).
+        input_tensors: optional list of input tensors
+            to build the model upon. If not provided,
+            placeholders will be created.
+
+    # Returns
+        An instance of `Model` reproducing the behavior
+        of the original model, on top of new inputs tensors,
+        using newly instantiated weights.
+
+    # Raises
+        ValueError: in case of invalid `model` argument value.
+    """
+    if isinstance(model, Sequential):
+        return _clone_sequential_model(model, input_tensors=input_tensors)
+    else:
+        return _clone_functional_model(model, input_tensors=input_tensors)
