@@ -24,7 +24,7 @@ class CustomObjectScope(object):
 
     # Example
 
-    Consider a custom object `MyObject`
+    Consider a custom object `MyObject` (e.g. a class):
 
     ```python
         with CustomObjectScope({'MyObject':MyObject}):
@@ -258,8 +258,6 @@ class Progbar(object):
 
     def __init__(self, target, width=30, verbose=1, interval=0.05):
         self.width = width
-        if target is None:
-            target = -1
         self.target = target
         self.sum_values = {}
         self.unique_values = []
@@ -269,6 +267,8 @@ class Progbar(object):
         self.total_width = 0
         self.seen_so_far = 0
         self.verbose = verbose
+        self._dynamic_display = (sys.stdout.isatty() or
+                                 'ipykernel' in sys.modules)
 
     def update(self, current, values=None, force=False):
         """Updates the progress bar.
@@ -291,18 +291,23 @@ class Progbar(object):
         self.seen_so_far = current
 
         now = time.time()
+        info = ' - %.0fs' % (now - self.start)
         if self.verbose == 1:
-            if not force and (now - self.last_update) < self.interval:
+            if (not force and (now - self.last_update) < self.interval and
+                    current < self.target):
                 return
 
             prev_total_width = self.total_width
-            sys.stdout.write('\b' * prev_total_width)
-            sys.stdout.write('\r')
+            if self._dynamic_display:
+                sys.stdout.write('\b' * prev_total_width)
+                sys.stdout.write('\r')
+            else:
+                sys.stdout.write('\n')
 
-            if self.target is not -1:
+            if self.target is not None:
                 numdigits = int(np.floor(np.log10(self.target))) + 1
-                barstr = '%%%dd/%%%dd [' % (numdigits, numdigits)
-                bar = barstr % (current, self.target)
+                barstr = '%%%dd/%d [' % (numdigits, self.target)
+                bar = barstr % current
                 prog = float(current) / self.target
                 prog_width = int(self.width * prog)
                 if prog_width > 0:
@@ -313,23 +318,39 @@ class Progbar(object):
                         bar += '='
                 bar += ('.' * (self.width - prog_width))
                 bar += ']'
-                sys.stdout.write(bar)
-                self.total_width = len(bar)
+            else:
+                bar = '%7d/Unknown' % current
+
+            self.total_width = len(bar)
+            sys.stdout.write(bar)
 
             if current:
                 time_per_unit = (now - self.start) / current
             else:
                 time_per_unit = 0
-            eta = time_per_unit * (self.target - current)
-            info = ''
-            if current < self.target and self.target is not -1:
-                info += ' - ETA: %ds' % eta
+            if self.target is not None and current < self.target:
+                eta = time_per_unit * (self.target - current)
+                if eta > 3600:
+                    eta_format = '%d:%02d:%02d' % (eta // 3600, (eta % 3600) // 60, eta % 60)
+                elif eta > 60:
+                    eta_format = '%d:%02d' % (eta // 60, eta % 60)
+                else:
+                    eta_format = '%ds' % eta
+
+                info = ' - ETA: %s' % eta_format
             else:
-                info += ' - %ds' % (now - self.start)
+                if time_per_unit >= 1:
+                    info += ' %.0fs/step' % time_per_unit
+                elif time_per_unit >= 1e-3:
+                    info += ' %.0fms/step' % (time_per_unit * 1e3)
+                else:
+                    info += ' %.0fus/step' % (time_per_unit * 1e6)
+
             for k in self.unique_values:
                 info += ' - %s:' % k
                 if isinstance(self.sum_values[k], list):
-                    avg = self.sum_values[k][0] / max(1, self.sum_values[k][1])
+                    avg = np.mean(
+                        self.sum_values[k][0] / max(1, self.sum_values[k][1]))
                     if abs(avg) > 1e-3:
                         info += ' %.4f' % avg
                     else:
@@ -339,25 +360,28 @@ class Progbar(object):
 
             self.total_width += len(info)
             if prev_total_width > self.total_width:
-                info += ((prev_total_width - self.total_width) * ' ')
+                info += (' ' * (prev_total_width - self.total_width))
+
+            if self.target is not None and current >= self.target:
+                info += '\n'
 
             sys.stdout.write(info)
             sys.stdout.flush()
 
-            if current >= self.target:
-                sys.stdout.write('\n')
-
-        if self.verbose == 2:
-            if current >= self.target:
-                info = '%ds' % (now - self.start)
+        elif self.verbose == 2:
+            if self.target is None or current >= self.target:
                 for k in self.unique_values:
                     info += ' - %s:' % k
-                    avg = self.sum_values[k][0] / max(1, self.sum_values[k][1])
+                    avg = np.mean(
+                        self.sum_values[k][0] / max(1, self.sum_values[k][1]))
                     if avg > 1e-3:
                         info += ' %.4f' % avg
                     else:
                         info += ' %.4e' % avg
-                sys.stdout.write(info + "\n")
+                info += '\n'
+
+                sys.stdout.write(info)
+                sys.stdout.flush()
 
         self.last_update = now
 
